{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "ef0fcfdd",
   "metadata": {},
   "source": [
    "# Week XI - Simulated Annealing"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "2ba562f9",
   "metadata": {},
   "source": [
    "In this tutorial we will code in Python a small example of 1D simulated annealing."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "cc2f0ade",
   "metadata": {},
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "import matplotlib.pyplot as plt\n",
    "from matplotlib import animation\n",
    "\n",
    "class Anneal():\n",
    "    '''\n",
    "    This class implements \n",
    "    simulated annealing for the\n",
    "    1D function (x+0.2)x+cos(14.5x-0.3)\n",
    "    '''\n",
    "    \n",
    "    def __init__(self,nsteps,T,T_eta,filename='traj.gif'):\n",
    "        'Init the class'\n",
    "        self.nsteps = nsteps\n",
    "        self.T = T\n",
    "        self.T_eta = T_eta\n",
    "        self.T_th = 1e-5\n",
    "        self.scale = 0.5\n",
    "        self.filename = filename\n",
    "        \n",
    "    @staticmethod\n",
    "    def f(x):\n",
    "        return (x+0.2)*x+np.cos(14.5*x-0.3)\n",
    "    \n",
    "    def init(self):\n",
    "        'Init the search'\n",
    "        self.rng = np.random.default_rng(seed=42424)\n",
    "        self.x = 1.0\n",
    "        self.y = self.f(self.x)\n",
    "        self.traj = []\n",
    "        self.y_min = self.y\n",
    "        self.x_min = self.x\n",
    "        \n",
    "    def run_step(self):\n",
    "        for i in range(self.nsteps):\n",
    "            rnd = self.rng.random(size=2)\n",
    "            x = self.x + self.scale*np.sqrt(self.T)*(rnd[0]-0.5)\n",
    "            y = self.f(x)\n",
    "            if (np.exp(-(y-self.y)/self.T)>rnd[1]):\n",
    "                self.y = y\n",
    "                self.x = x\n",
    "                self.traj.append((x,y))\n",
    "            if (self.y<self.y_min):\n",
    "                self.x_min = self.x\n",
    "                self.y_min = self.y\n",
    "                \n",
    "    def run_anneal(self):\n",
    "        while (self.T>=self.T_th):\n",
    "            self.run_step()\n",
    "            self.T = self.T * self.T_eta\n",
    "            \n",
    "    def plot(self):\n",
    "        x = np.linspace(-3,3,1000)\n",
    "        y = map(self.f,x)\n",
    "        plt.xlabel('x')\n",
    "        plt.ylabel('f(x)')\n",
    "        plt.xlim([-3,3])\n",
    "        plt.ylim([-2,10])\n",
    "        plt.plot(x,list(y),'-',label = '$(x+0.2)x+\\cos{(14.5x-0.3)}$')\n",
    "        t = np.array(self.traj)\n",
    "        plt.scatter(t[:,0],t[:,1],s=5,c='red',label = 'trajectory')\n",
    "        plt.scatter(self.x_min,self.y_min,s=100,c='green',marker='*',label = 'minimum')\n",
    "        plt.legend()\n",
    "        plt.show()\n",
    "        plt.close()\n",
    "    \n",
    "    def animate(self):\n",
    "        fig = plt.figure() \n",
    "        axis = plt.axes(xlim =(-3, 3),\n",
    "                ylim =(-1, 4)) \n",
    "  \n",
    "        line, = axis.plot([], [], marker='o',ls='',c='red',ms=4,lw=2) \n",
    "        xx = np.linspace(-3,3,1000)\n",
    "        yy = map(self.f,xx)\n",
    "        plt.plot(xx,list(yy),'-',c='k',label = '$(x+0.2)x+\\cos{(14.5x-0.3)}$')\n",
    "        plt.xlabel('x')\n",
    "        plt.ylabel('f(x)')\n",
    "        def init(): \n",
    "            line.set_data([], []) \n",
    "            return line, \n",
    "        xdata, ydata = [], [] \n",
    "        def animate(i): \n",
    "            #axis.clear()\n",
    "            axis.set_xlim([-3,3])\n",
    "            axis.set_ylim([-1,3])\n",
    "            x = self.traj[i][0]\n",
    "            y = self.traj[i][1]\n",
    "    # appending values to the previously \n",
    "    # empty x and y data holders \n",
    "            xdata.append(x) \n",
    "            ydata.append(y) \n",
    "            line.set_data(xdata, ydata) \n",
    "         \n",
    "            return line,\n",
    "   \n",
    "        anim = animation.FuncAnimation(fig, animate, init_func = init, \n",
    "                               frames = len(self.traj), interval = 20,blit=True) \n",
    "        anim.save(self.filename, writer = 'pillow', fps = 30)\n",
    "        plt.close()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "c07b38d1",
   "metadata": {},
   "source": [
    "Let us plot the cost function: several local minima are clearly visibile!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "37c3119e",
   "metadata": {},
   "outputs": [],
   "source": [
    "x = np.linspace(-3,3,1000)\n",
    "y = map(Anneal.f,x)\n",
    "plt.xlabel('x')\n",
    "plt.ylabel('f(x)')\n",
    "plt.xlim([-3,3])\n",
    "plt.plot(x,list(y),'-',label = '$(x+0.2)x+\\cos{(14.5x-0.3)}$')\n",
    "plt.legend()\n",
    "plt.show()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "aa0e820f",
   "metadata": {},
   "source": [
    "Let's run the code!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "37d1361d",
   "metadata": {},
   "outputs": [],
   "source": [
    "# 10 steps, T=1 and T_eta = 0.1\n",
    "anneal = Anneal(10,1,0.1,'traj1.gif')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "6fee8c16",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Initialization of the search\n",
    "anneal.init()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "7406e1c1",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Simulated annealing\n",
    "anneal.run_anneal()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "69800117",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Visualization\n",
    "anneal.plot()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "8c0e01d9",
   "metadata": {},
   "source": [
    "Now let's look at what happened during the simulated annealing."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d7a4d5a5",
   "metadata": {},
   "outputs": [],
   "source": [
    "_ = anneal.animate()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3440be1b",
   "metadata": {},
   "outputs": [],
   "source": [
    "from IPython.display import HTML\n",
    "HTML('<img src=\"./traj1.gif\">')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "dc3fa3f3",
   "metadata": {},
   "outputs": [],
   "source": [
    "#<img src=\"traj.gif\" width=\"750\" align=\"center\">"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "d07ae44c",
   "metadata": {},
   "source": [
    "We did not reach the global minimum.\n",
    "\n",
    "Now we increase the temperature and slow down the annealing."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "f8c7d6b2",
   "metadata": {},
   "outputs": [],
   "source": [
    "anneal = Anneal(10,2,0.97,'traj2.gif')\n",
    "anneal.init()\n",
    "anneal.run_anneal()\n",
    "anneal.plot()\n",
    "_ = anneal.animate()\n",
    "HTML('<img src=\"./traj2.gif\">')"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e69ff633",
   "metadata": {},
   "source": [
    "We found the minimum!\n",
    "\n",
    "How would it be if we try more MC steps and higher temperature?"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "564e0a2d",
   "metadata": {},
   "outputs": [],
   "source": [
    "anneal = Anneal(4000,10,0.95)\n",
    "anneal.init()\n",
    "anneal.run_anneal()\n",
    "anneal.plot()\n",
    "print('minimum x={}, f(x)={}'.format(anneal.x_min,anneal.y_min))\n",
    "print('final x={}, f(x)={}'.format(anneal.x,anneal.y))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "c3325b6a",
   "metadata": {},
   "outputs": [],
   "source": [
    "anneal = Anneal(4000,1,0.95,'traj_test.gif')\n",
    "anneal.init()\n",
    "anneal.run_anneal()\n",
    "anneal.plot()\n",
    "print('minimum x={}, f(x)={}'.format(anneal.x_min,anneal.y_min))\n",
    "print('final x={}, f(x)={}'.format(anneal.x,anneal.y))"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "569a6e94",
   "metadata": {},
   "source": [
    "We have seen how the performance of simulated annealing crucially depends on the annealing temperature, the cooling schedule and the number of Monte Carlo steps."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e99a70dc",
   "metadata": {},
   "source": [
    "---------------------------"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "18e8898b",
   "metadata": {},
   "source": [
    "## The Traveling Salesman problem: some hints for Homework 2\n",
    "\n",
    "In Homework 2, you will have to use simulated annealing to solve the traveling salesman problem\n",
    "\n",
    "Problem 15.31a from *An Introduction to Computer Simulation Methods Third Edition (revised)* by Harvey Gould, Jan Tobochnik, and Wolfgang Christian.\n",
    "> Generate a random arrangement of N = 8 cities in a square of linear dimension L = $\\sqrt{N}$\n",
    "> and calculate the optimum route by hand. Then write a Monte Carlo program and apply \n",
    "> the method of simulated annealing to this problem. For example, use two arrays to store \n",
    "> the x and y coordinate of each city and an array to store the distances between them. \n",
    "> The state of the system, that is, the route represented by a sequence of cities, can be stored in another array. \n",
    "> The length of this route is associated with the energy of an imaginary thermal system. \n",
    "> A reasonable choice for the initial temperature is one that is the same order as the initial energy. \n",
    "> One way to generate a random rearrangement of the route is to choose two cities at random and to\n",
    "> interchange the order of visit. Choose this method or one that you devise and find a reasonable annealing schedule.\n",
    "> Compare your annealing results to exact results whenever possible. Extend your results to larger N, for example, N = 12, 24, and 48.\n",
    "> For a given annealing schedule, determine the probability of finding a route of a given length. More suggestions can be found in the references.\n",
    "\n",
    "You will have to write the code, here below you find only some hints to start working on it..."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "ee32a617",
   "metadata": {},
   "source": [
    "*CODE REMOVED*"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "fis_comp_2023",
   "language": "python",
   "name": "fis_comp_2023"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.1"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
